/* BBMediator - A BB4W plugin to execute something on some message
   Copyright (C) 2004-2005 kana <nicht AT s8 DOT xrea DOT con>

This program is released under the terms of GNU General Public License;
either version 2 of the license or (at your option) any later version.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBApi-C.h>
#include <BBCmd.h>

#include "CommandTable.h"
#include "MessageTable.h"
#include "misc.h"




static const char BBMEDIATOR_NAME[] = "BBMediator";
static const char BBMEDIATOR_VERSION[] = "0.0.1";
static const char BBMEDIATOR_AUTHOR[] = "kana";
static const char BBMEDIATOR_RELEASE[] = "2004-06-07/2005-01-09";
static const char BBMEDIATOR_LINK[]
                      = "http://nicht.s8.xrea.com/2004/06/07/BBMediator";
static const char BBMEDIATOR_EMAIL[] = "nicht AT s8 DOT xrea DOT con";




static HWND G_core_cache = NULL;
static HHOOK G_hook_callwndproc = NULL;
static HHOOK G_hook_callwndprocret = NULL;
static HHOOK G_hook_getmessage = NULL;
static CommandTable* G_command_table = NULL;








void
notice(const char* format, ...)
{
	char buf[80 * 25];
	va_list va;

	va_start(va, format);
	vsnprintf(buf, NUMBER_OF(buf), format, va);
	va_end(va);

	MessageBox( NULL, buf, BBMEDIATOR_NAME,
	            MB_OK | MB_SETFOREGROUND | MB_TASKMODAL | MB_TOPMOST );
}




FILE*
open_conf_file(HINSTANCE hinstance, const char* filename)
{
	char buf[MAX_PATH];
	char dir[MAX_PATH];
	char* p;

	if (!GetModuleFileName(hinstance, dir, NUMBER_OF(dir)))
		return NULL;
	dir[NUMBER_OF(dir) - 1] = '\0';

	p = dir + strlen(dir);
	while (dir <= p) {
		if (*p == '\\') {p[1] = '\0';  break;}
		p--;
	}

	snprintf(buf, NUMBER_OF(buf), "%s%s", dir, filename);
	buf[NUMBER_OF(buf) - 1] = '\0';

	return fopen(buf, "rt");
}








static void
execute(UINT msg, _Timing timing, WPARAM wp, LPARAM lp)
{
	CommandTable_LookUpAndExecute(
		G_command_table, msg, timing, wp, lp,
		CT_DIR_FIRST_TO_LAST, 0
	);
}




static LRESULT CALLBACK
hook_proc_callwndproc(int ncode, WPARAM wp, LPARAM lp)
{
	CWPSTRUCT* data;

	data = (CWPSTRUCT*)lp;

	if (ncode == HC_ACTION && data->hwnd == G_core_cache) {
		execute( data->message, TIMING_BEFORE,
		         data->wParam, data->lParam );
	}

	return CallNextHookEx(NULL, ncode, wp, lp);
}




static LRESULT CALLBACK
hook_proc_callwndprocret(int ncode, WPARAM wp, LPARAM lp)
{
	CWPRETSTRUCT* data;

	data = (CWPRETSTRUCT*)lp;

	if (ncode == HC_ACTION && data->hwnd == G_core_cache) {
		execute( data->message, TIMING_AFTER,
		         data->wParam, data->lParam );
	}

	return CallNextHookEx(NULL, ncode, wp, lp);
}




static LRESULT CALLBACK
hook_proc_getmessage(int ncode, WPARAM wp, LPARAM lp)
{
	MSG* msg;

	msg = (MSG*)lp;

	if (ncode==HC_ACTION && wp==PM_REMOVE && msg->hwnd==G_core_cache) {
		execute( msg->message, TIMING_BEFORE,
		         msg->wParam, msg->lParam );
	}

	return CallNextHookEx(NULL, ncode, wp, lp);
}








DLL_EXPORT int
beginPlugin(HINSTANCE hplugininstance)
{
	FILE* file;
	MessageTable* message_table;

	if ( BBCmd_IsCompatible(BBCmd_VERSION, BBCmd_ICH_JUST_USE)
	     != BBCmd_IC_COMPATIBLE )
	{
		notice("Incompatible BBCmd");
		goto E_BBCMD_COMPATIBILITY;
	}

	file = open_conf_file(hplugininstance, "BBMediator.msgs");
	if (file == NULL)
		{notice("BBMediator.msgs not found");  goto E_LOAD_MSGS_O;}
	message_table = MessageTable_FromFile(file);
	fclose(file);
	if (message_table == NULL)
		{notice("Failed to load BBMediator.msgs");  goto E_LOAD_MSGS;}


	file = open_conf_file(hplugininstance, "BBMediator.cmds");
	if (file == NULL)
		{notice("BBMediator.cmds not found");  goto E_LOAD_CMDS_O;}
	G_command_table = CommandTable_FromFile(file, message_table);
	fclose(file);
	if (G_command_table == NULL)
		{notice("Failed to load BBMediator.cmds");  goto E_LOAD_CMDS;}

	MessageTable_Del(message_table);
	message_table = NULL;




	G_core_cache = GetBBWnd();

	G_hook_callwndproc
		= SetWindowsHookEx( WH_CALLWNDPROC, hook_proc_callwndproc,
		                    NULL, GetCurrentThreadId() );
	if (G_hook_callwndproc == NULL) {
		notice("Failed to hook WH_CALLWNDPROC");
		goto E_HOOK_CALLWNDPROC;
	}

	G_hook_callwndprocret
		= SetWindowsHookEx( WH_CALLWNDPROCRET,hook_proc_callwndprocret,
		                    NULL, GetCurrentThreadId() );
	if (G_hook_callwndprocret == NULL) {
		notice("Failed to hook WH_CALLWNDPROCRET");
		goto E_HOOK_CALLWNDPROCRET;
	}

	G_hook_getmessage
		= SetWindowsHookEx( WH_GETMESSAGE, hook_proc_getmessage,
		                    NULL, GetCurrentThreadId() );
	if (G_hook_getmessage == NULL) {
		notice("Failed to hook WH_GETMESSAGE");
		goto E_HOOK_GETMESSAGE;
	}


	SendMessage( G_core_cache, BB_BROADCAST, 0,
	             (LPARAM)"@BBMediator.loaded" );

	return 0;


E_HOOK_GETMESSAGE:
	UnhookWindowsHookEx(G_hook_callwndprocret);
E_HOOK_CALLWNDPROCRET:
	UnhookWindowsHookEx(G_hook_callwndproc);
E_HOOK_CALLWNDPROC:
	CommandTable_Del(G_command_table);
E_LOAD_CMDS:
E_LOAD_CMDS_O:
	if (message_table != NULL) MessageTable_Del(message_table);
E_LOAD_MSGS:
E_LOAD_MSGS_O:
E_BBCMD_COMPATIBILITY:
	return !0;
}




DLL_EXPORT void
endPlugin(HINSTANCE hplugininstance)
{
	SendMessage( G_core_cache, BB_BROADCAST, 0,
	             (LPARAM)"@BBMediator.unloaded" );

	UnhookWindowsHookEx(G_hook_getmessage);
	UnhookWindowsHookEx(G_hook_callwndprocret);
	UnhookWindowsHookEx(G_hook_callwndproc);
	CommandTable_Del(G_command_table);
}




DLL_EXPORT LPCSTR
pluginInfo(int field)
{
	switch (field) {
		case PLUGIN_NAME:    return BBMEDIATOR_NAME;
		case PLUGIN_VERSION: return BBMEDIATOR_VERSION;
		case PLUGIN_AUTHOR:  return BBMEDIATOR_AUTHOR;
		case PLUGIN_RELEASE: return BBMEDIATOR_RELEASE;
		case PLUGIN_LINK:    return BBMEDIATOR_LINK;
		case PLUGIN_EMAIL:   return BBMEDIATOR_EMAIL;
		default:             return BBMEDIATOR_NAME;
	}
}




/* __END__ */
