/* Data structure and interface for command table (BBMediator.cmds)

FIXME: Dirty!
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <BBCmd.h>

#include "CommandTable.h"
#include "MessageTable.h"
#include "Pattern.h"

#include "misc.h"




typedef struct {
	UINT msg;
	_Timing timing;
	Pattern* pattern_wp;
	Pattern* pattern_lp;
	BBCmdCommand* script;
} CommandTableEntry;


struct _CommandTable {
	CommandTableEntry** entries;
	int num;
};








static CommandTableEntry*
CommandTableEntry_New( UINT msg, _Timing timing, Pattern* pattern_wp,
                       Pattern* pattern_lp, BBCmdCommand* script )
{
	CommandTableEntry* self;

	self = malloc(sizeof(*self));
	if (self == NULL) return NULL;

	self->msg = msg;
	self->timing = timing;
	self->pattern_wp = pattern_wp;
	self->pattern_lp = pattern_lp;
	self->script = script;

	return self;
}


static void
CommandTableEntry_Del(CommandTableEntry* self)
{
	Pattern_Del(self->pattern_wp);
	Pattern_Del(self->pattern_lp);
	BBCmdCommand_Del(self->script);
	free(self);
}








static int
_CommandTable_AddEntry( CommandTable* self,
                        UINT msg, _Timing timing, Pattern* pattern_wp,
                        Pattern* pattern_lp, BBCmdCommand* script )
{
	CommandTableEntry** tmp;
	CommandTableEntry* new_entry;

	new_entry = CommandTableEntry_New( msg, timing, pattern_wp,
	                                   pattern_lp, script );
	if (new_entry == NULL) {
		Pattern_Del(pattern_wp);
		Pattern_Del(pattern_lp);
		BBCmdCommand_Del(script);
		goto E_ALLOCATE_NEW_ENTRY;
	}

	tmp = realloc(self->entries, (self->num + 1) * sizeof(*tmp));
	if (tmp == NULL) goto E_EXPAND_TABLE;

	self->entries = tmp;
	self->entries[self->num] = new_entry;
	self->num++;

	return 0;


E_EXPAND_TABLE:
	CommandTableEntry_Del(new_entry);
E_ALLOCATE_NEW_ENTRY:
	return !0;
}


CommandTable*
CommandTable_FromFile(FILE* file, const MessageTable* message_table)
{
	CommandTable* self;
	char buf[256];
	int lineno;

	BBCmdCompiler* compiler;
	int e_line;
	int e_column;
	const char* e_msg;


	compiler = BBCmdCompiler_New();
	if (compiler == NULL) {
		notice(".cmds: Out of memory to make BBCmd compiler");
		goto E_MAKE_COMPILER;
	}

	self = malloc(sizeof(*self));
	if (self == NULL) {
		notice(".cmds: Out of memory to allocate table");
		goto E_ALLOC_TABLE;
	}

	self->num = 0;
	self->entries = NULL;

	lineno = 0;

	while (fgets(buf, NUMBER_OF(buf), file) != NULL) {
		char* line;
		char msg_name[NUMBER_OF(buf)];
		char timing_c;
		char wp_s[NUMBER_OF(buf)];
		char lp_s[NUMBER_OF(buf)];
		char script_src[NUMBER_OF(buf)];
		UINT msg;
		_Timing timing;
		Pattern* pattern_wp;
		Pattern* pattern_lp;
		BBCmdCommand* script;

		lineno++;

		line = buf;
		while (*line == ' ' || *line == '\t')
			line++;
		if (*line == '#' || *line == '\n')
			continue;

		/*                 msg  timing wp      lp      script */
		if (sscanf( line, "%[^\t]\t%c\t%[^\t]\t%[^\t]\t%[^\t]",
		            msg_name, &timing_c, wp_s, lp_s, script_src ) != 5)
		{
			notice(".cmds: Syntax error at line %d", lineno);
			continue;
		}


		msg = MessageTable_LookUpValueByName(message_table, msg_name);
		if (msg == 0) {
			notice( ".cmds: Invalid message name `%s' at line %d",
			        msg_name, lineno );
			goto E_MSG;
		}

		switch (timing_c) {
		default:
			notice( ".cmds: Invalid timing `%c' at line %d",
			        timing_c, lineno );
			goto E_TIMING;

		case 'a': case 'A': timing = TIMING_AFTER; break;

		case '-': /* FALLTHRU */
		case 'b': case 'B': timing = TIMING_BEFORE; break;
		}

		pattern_wp = Pattern_Compile(wp_s);
		if (pattern_wp == NULL) {
			notice( ".cmds: Syntax error about wp at line %d",
			        lineno );
			goto E_WP;
		}

		pattern_lp = Pattern_Compile(lp_s);
		if (pattern_lp == NULL) {
			notice( ".cmds: Syntax error about lp at line %d",
			        lineno );
			goto E_LP;
		}

		script = BBCmdCompiler_CompileFromString
		           (compiler, script_src, &e_line, &e_column, &e_msg);
		if (script == NULL) {
			notice( ".cmds: BBCmd script compiling error (%s)"
			        " at line %d", e_msg, lineno );
			goto E_SCRIPT;
		}

		if (_CommandTable_AddEntry( self, msg, timing, pattern_wp,
		                            pattern_lp, script ) != 0)
		{
			notice(".cmds: Out of memory at %d", lineno);
			goto E_ADD_ENTRY;
		}

		continue;

	E_SCRIPT:
		Pattern_Del(pattern_lp);
	E_LP:
		Pattern_Del(pattern_wp);
	E_WP:
	E_TIMING:
	E_MSG:
		continue;
	}

	BBCmdCompiler_Del(compiler);

	return self;


E_ADD_ENTRY:
	CommandTable_Del(self);
E_ALLOC_TABLE:
	BBCmdCompiler_Del(compiler);
E_MAKE_COMPILER:
	return NULL;
}


void
CommandTable_Del(CommandTable* self)
{
	int i;

	for (i = 0; i < self->num; i++)
		CommandTableEntry_Del(self->entries[i]);

	free(self->entries);
	free(self);
}




void
CommandTable_LookUpAndExecute( const CommandTable* self,
                               UINT msg, _Timing timing, WPARAM wp, LPARAM lp,
                               _CT_Dir dir, int threshold )
{
	int begin;
	int end;
	int diff;
	int i;
	int t;

	if (dir == CT_DIR_FIRST_TO_LAST) {
		begin = 0;
		end = self->num;
		diff = 1;
	} else { /* CT_DIR_LAST_TO_FIRST */
		begin = self->num - 1;
		end = -1;
		diff = -1;
	}

	t = threshold;

	for (i = begin; i != end; i += diff) {
		CommandTableEntry* e;

		e = self->entries[i];

		if ( msg == e->msg && timing == e->timing
		     && Pattern_MatchWparam(e->pattern_wp, wp)
		     && Pattern_MatchLparam(e->pattern_lp, lp) )
		{
			BBCmdCommand_Execute(e->script);

			if (threshold) {
				t--;

				if (!t) break;
			}
		}
	}

	return;
}




/* __END__ */
