/* Data structure and interface for message table (BBMediator.msgs) */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include "MessageTable.h"

#include "misc.h"




typedef struct {
	UINT value;
	char name[1];
} MessageTableEntry;


struct _MessageTable {
	MessageTableEntry** entries;
	int num;
};








static MessageTableEntry*
MessageTableEntry_New(const char* name, UINT value)
{
	MessageTableEntry* self;

	self = malloc(sizeof(*self) + strlen(name) * sizeof(char));
	if (self == NULL) return NULL;

	self->value = value;
	strcpy(self->name, name);

	return self;
}


static void
MessageTableEntry_Del(MessageTableEntry* self)
{
	free(self);
}








static int
_MessageTable_AddEntry(MessageTable* self, const char* name, UINT value)
{
	MessageTableEntry** tmp;
	MessageTableEntry* new_entry;

	new_entry = MessageTableEntry_New(name, value);
	if (new_entry == NULL) goto E_ALLOCATE_NEW_ENTRY;

	tmp = realloc(self->entries, (self->num + 1) * sizeof(*tmp));
	if (tmp == NULL) goto E_EXPAND_TABLE;

	self->entries = tmp;
	self->entries[self->num] = new_entry;
	self->num++;

	return 0;


E_EXPAND_TABLE:
	MessageTableEntry_Del(new_entry);
E_ALLOCATE_NEW_ENTRY:
	return !0;
}


MessageTable*
MessageTable_FromFile(FILE* file)
{
	MessageTable* self;
	char buf[256];
	char name[NUMBER_OF(buf)];
	UINT value;
	int lineno;

	self = malloc(sizeof(*self));
	if (self == NULL) return NULL;

	self->num = 0;
	self->entries = NULL;

	lineno = 0;

	while (fgets(buf, NUMBER_OF(buf), file) != NULL) {
		char* line;

		lineno++;

		line = buf;
		while (*line == ' ' || *line == '\t')
			line++;
		if (*line == '#' || *line == '\n')
			continue;

		if (sscanf(line, "%s %u\n", name, &value) != 2) {
			notice(".msgs: Syntax error at line %d", lineno);
			continue;
		}

		if (_MessageTable_AddEntry(self, name, value) != 0) {
			notice(".msgs: Out of memory at line %d", lineno);
			goto E_ADD_ENTRY;
		}
	}

	return self;


E_ADD_ENTRY:
	MessageTable_Del(self);
	return NULL;
}


void
MessageTable_Del(MessageTable* self)
{
	int i;

	for (i = 0; i < self->num; i++)
		MessageTableEntry_Del(self->entries[i]);

	free(self->entries);
	free(self);
}




UINT
MessageTable_LookUpValueByName(const MessageTable* self, const char* name)
{
	int i;

	for (i = 0; i < self->num; i++)
		if (!strcmp(name, self->entries[i]->name))
			return self->entries[i]->value;

	return 0;
}




/* __END__ */
